/*
 * GStreamer
 * Copyright (C) 2006 Stefan Kost <ensonic@users.sf.net>
 * Copyright (C) 2014 Enrico Scholz <enrico.scholz@sigma-chemnitz.de>
 * Copyright (C) 2018 PHYTEC Messtechnik GmbH
 * Author: Stefan Riedmueller <s.riedmueller@phytec.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-vita1300_remapper
 *
 * FIXME:Describe vita1300_remapper here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! vita1300_remapper ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdbool.h>

#include <gst/gst.h>
#include <gst/base/base.h>
#include <gst/controller/controller.h>
#include <gst/video/video.h>

#include <vita-1300_remapper/vita1300-remapper.h>

#include "gstvita-1300remapper.h"

GST_DEBUG_CATEGORY_STATIC(gst_vita1300remapper_debug);
#define GST_CAT_DEFAULT gst_vita1300remapper_debug

/* Filter signals and args */
enum {
	/* FILL ME */
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_SILENT,
	PROP_MODE,
	PROP_PT,
};

#define CAP_BAYER(_fmt) \
	"video/x-bayer, "			      \
	"format = " _fmt ", "			      \
	"width=[ 16, 2147483647 ], "		      \
	"height = " GST_VIDEO_SIZE_RANGE ", "	      \
        "framerate = " GST_VIDEO_FPS_RANGE

#define V1300_CAPS					\
	GST_VIDEO_CAPS_MAKE("{ GRAY8, GRAY16_BE, GRAY16_LE, "		\
			    "RGB15, RGB16, xRGB, RGBx, xBGR, BGRx, "	\
			    "RGB, BGR }") ";"				\
	CAP_BAYER("(string){ bggr, rggb, grbg, gbrg }") ";"	\

/* the capabilities of the inputs and outputs.
 *
 * FIXME:describe the real formats here.
 */
static GstStaticPadTemplate	sink_template = {
	.name_template = "sink",
	.direction = GST_PAD_SINK,
	.presence = GST_PAD_ALWAYS,
	.static_caps = GST_STATIC_CAPS(V1300_CAPS),
};

static GstStaticPadTemplate	src_template = {
	.name_template = "src",
	.direction = GST_PAD_SRC,
	.presence = GST_PAD_ALWAYS,
	.static_caps = GST_STATIC_CAPS(V1300_CAPS),
};

typedef struct vita1300_remapper vita1300_remapper;
#define vita1300_remapperClass	struct vita1300_remapper_class
#define gst_vita1300_remapper_parent_class	parent_class
G_DEFINE_TYPE(vita1300_remapper, gst_vita1300_remapper, GST_TYPE_BASE_TRANSFORM);
#undef vita1300_remapperClass

static void
gst_vita1300remapper_set_property(GObject *object, guint prop_id,
				  const GValue *value, GParamSpec *pspec)
{
	struct vita1300_remapper *filter = GST_VITA1300_REMAPPER(object);

	switch (prop_id) {
	case PROP_SILENT:
		filter->silent = g_value_get_boolean(value);
		break;
	case PROP_MODE:
		filter->mode = g_value_get_int(value);
		break;
	case PROP_PT:
		gst_base_transform_set_passthrough(&filter->element,
						   g_value_get_boolean(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}
}

static void
gst_vita1300remapper_get_property(GObject *object, guint prop_id,
				  GValue *value, GParamSpec *pspec)
{
	struct vita1300_remapper *filter = GST_VITA1300_REMAPPER(object);

	switch (prop_id) {
	case PROP_SILENT:
		g_value_set_boolean(value, filter->silent);
		break;
	case PROP_MODE:
		g_value_set_int(value, filter->mode);
		break;
	case PROP_PT:
		g_value_set_boolean(value,
				    gst_base_transform_is_passthrough(&filter->element));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}
}

/* GstBaseTransform vmethod implementations */

/* this function does the actual processing
 */
static GstFlowReturn
gst_vita1300remapper_transform(GstBaseTransform *base, GstBuffer *inbuf,
			       GstBuffer *outbuf)
{
	struct vita1300_remapper	*filter = GST_VITA1300_REMAPPER(base);
	GstBaseTransform		*btrans = GST_BASE_TRANSFORM(filter);

	GstMapInfo			inmap_info;
	GstMapInfo			outmap_info;

	enum vita1300_result		res;
	int				bpp;

	if (GST_CLOCK_TIME_IS_VALID(GST_BUFFER_TIMESTAMP(outbuf)))
		gst_object_sync_values(GST_OBJECT(filter),
				       GST_BUFFER_TIMESTAMP(outbuf));

	if (gst_base_transform_is_passthrough (btrans))
		return GST_FLOW_OK;

	if (!gst_buffer_map(inbuf, &inmap_info, GST_MAP_READ)) {
		GST_ERROR_OBJECT(filter, "failed to map inbuf");
		return GST_FLOW_ERROR;
	}

	if (!gst_buffer_map(outbuf, &outmap_info, GST_MAP_WRITE)) {
		GST_ERROR_OBJECT(filter, "failed to map outbuf");
		gst_buffer_unmap(inbuf, &inmap_info);
		return GST_FLOW_ERROR;
	}

	if (inmap_info.size != outmap_info.size) {
		GST_ERROR_OBJECT(filter,
				 "mismatch of input and output buffer sizes (%zu vs. %zu)",
				 inmap_info.size, outmap_info.size);
		gst_buffer_unmap(outbuf, &outmap_info);
		gst_buffer_unmap(inbuf,  &inmap_info);
		return GST_FLOW_ERROR;
	}

	bpp = filter->bpp;

	res = vita1300_convert(outmap_info.data, inmap_info.data, filter->mode,
			       filter->width, filter->height, bpp);

	GST_DEBUG_OBJECT(filter, "converting %p->%p (%dx%d@%d) in mode %d -> %d",
			 inmap_info.data, outmap_info.data,
			 filter->width, filter->height, bpp, filter->mode, res);

	gst_buffer_unmap(outbuf, &outmap_info);
	gst_buffer_unmap(inbuf, &inmap_info);

	switch (res) {
	case VITA1300_RES_OK:
		return GST_FLOW_OK;

	case VITA1300_BAD_X:
		GST_ERROR_OBJECT (filter, "invalid width: %d", filter->width);
		break;

	case VITA1300_BAD_MODE:
		GST_ERROR_OBJECT (filter, "invalid mode: %d", filter->mode);
		break;

	case VITA1300_BAD_BPP:
		GST_ERROR_OBJECT (filter, "unsupported bpp: %d", bpp);
		break;

	case VITA1300_BAD_OVERLAP:
		GST_ERROR_OBJECT (filter, "overlap error");
		break;
	}

	return GST_FLOW_ERROR;
}

enum bayer_format {
	BAYER_BGGR,
	BAYER_GRBG,
	BAYER_RGGB,
	BAYER_GBRG
};

static bool str_to_bayer(char const *str, enum bayer_format *format)
{
	if (strcmp(str, "bggr") == 0)
		*format = BAYER_BGGR;
	else if (strcmp(str, "grbg") == 0)
		*format = BAYER_GRBG;
	else if (strcmp(str, "rggb") == 0)
		*format = BAYER_RGGB;
	else if (strcmp(str, "gbrg") == 0)
		*format = BAYER_GBRG;
	else
		goto err;

	return true;

err:
	GST_ERROR("bad bayer format '%s'", str);
	return false;
}

static gboolean
gst_vita1300remapper_set_caps(GstBaseTransform *base, GstCaps *incaps,
			      GstCaps *outcaps)
{
	struct vita1300_remapper *filter = GST_VITA1300_REMAPPER(base);
	GstVideoInfo		info;
	GstStructure		*structure;
	guint			tmp;
	unsigned int		default_bpp = 8;

	GST_DEBUG_OBJECT (filter,
			  "setting caps: in %" GST_PTR_FORMAT
			  " out %" GST_PTR_FORMAT, incaps, outcaps);

	if (!gst_video_info_from_caps(&info, incaps)) {
		GST_ERROR_OBJECT(filter, "invalid input caps: %"
				 GST_PTR_FORMAT, incaps);
		return FALSE;
	}

	structure  = gst_caps_get_structure(incaps, 0);

	if (gst_structure_has_name(structure, "video/x-bayer")) {
		enum bayer_format	bayer_fmt;

		if (!str_to_bayer(gst_structure_get_string(structure, "format"),
				 &bayer_fmt))
			goto invalid_caps;

		default_bpp = 8;
	} else if (!info.finfo) {
		goto invalid_caps;
	} else {
		switch (info.finfo->format) {
		case GST_VIDEO_FORMAT_GRAY8:
		case GST_VIDEO_FORMAT_GRAY16_BE:
		case GST_VIDEO_FORMAT_GRAY16_LE:
			break;

		case GST_VIDEO_FORMAT_UNKNOWN:
			goto invalid_caps;

		default:
			break;
		}

		default_bpp = info.finfo->bits;
	}

	filter->width  = info.width;
	filter->height = info.height;

	if (gst_structure_get_uint(structure, "bpp", &tmp))
		filter->bpp = tmp;
	else
		filter->bpp = default_bpp;

	if (filter->bpp == 0) {
		GST_ERROR_OBJECT (filter, "Invalid bpp %d: %" GST_PTR_FORMAT,
				  filter->bpp, incaps);
		goto invalid_caps;
	}

	filter->stride = info.stride[0] * default_bpp / filter->bpp;

	return TRUE;

invalid_caps:
	GST_ERROR_OBJECT (filter, "Invalid caps: %" GST_PTR_FORMAT, incaps);
	return FALSE;
}

/* GObject vmethod implementations */

/* initialize the vita1300_remapper's class */
static void gst_vita1300_remapper_class_init(struct vita1300_remapper_class *klass)
{
	GstBaseTransformClass	*ktrans = GST_BASE_TRANSFORM_CLASS(klass);
	GstElementClass		*element_class = GST_ELEMENT_CLASS(klass);
	GObjectClass		*gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = gst_vita1300remapper_set_property;
	gobject_class->get_property = gst_vita1300remapper_get_property;

	g_object_class_install_property(
		gobject_class, PROP_SILENT,
		g_param_spec_boolean("silent", "Silent",
				     "Produce verbose output ?",
				     FALSE,
				     G_PARAM_READWRITE |
				     GST_PARAM_CONTROLLABLE));

	g_object_class_install_property(
		gobject_class, PROP_MODE,
		g_param_spec_int("mode", "Mode",
				 "remapping mode",
				 VITA1300_SUBSAMPLING_NONE,
				 VITA1300_SUBSAMPLING_COLOR,
				 VITA1300_SUBSAMPLING_NONE, G_PARAM_READWRITE));

	g_object_class_install_property(
		gobject_class, PROP_PT,
		g_param_spec_boolean("passthrough", "flag",
				     "sets passthrough mode",
				     FALSE,
				     G_PARAM_READWRITE |
				     GST_PARAM_CONTROLLABLE));

	gst_element_class_set_details_simple(
		element_class,
		"vita1300remapper",
		"Generic/Filter",
		"VITA 1300 Remapper",
		"Enrico Scholz <enrico.scholz@sigma-chemnitz.de>");

	gst_element_class_add_pad_template(
		element_class, gst_static_pad_template_get(&src_template));
	gst_element_class_add_pad_template(
		element_class, gst_static_pad_template_get(&sink_template));

	ktrans->transform =
		GST_DEBUG_FUNCPTR(gst_vita1300remapper_transform);
	ktrans->set_caps =
		GST_DEBUG_FUNCPTR(gst_vita1300remapper_set_caps);

	GST_DEBUG_CATEGORY_INIT(gst_vita1300remapper_debug, "vita1300remapper", 0,
				"vita1300remapper");
}

/* initialize the new element
 * initialize instance structure
 */
static void
gst_vita1300_remapper_init(struct vita1300_remapper *filter)
{
	filter->silent = FALSE;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean vita1300_remapper_init(GstPlugin *vita1300_remapper)
{
	return gst_element_register(vita1300_remapper, "vita1300_remapper",
				    GST_RANK_NONE, GST_TYPE_VITA1300_REMAPPER);
}

/* gstreamer looks for this structure to register vita1300_remappers
 *
 * FIXME:exchange the string 'Template vita1300_remapper' with you
 * vita1300_remapper description
 */
GST_PLUGIN_DEFINE(GST_VERSION_MAJOR,
		  GST_VERSION_MINOR,
		  vita_1300_remapper,
		  "VITA-1300 Remapper",
		  vita1300_remapper_init,
		  VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
